@@ -1,6 +1,7 @@
 import asyncio
 import math
 import warnings
+import logging
 from abc import ABC
 from operator import attrgetter
 from typing import (
@@ -26,17 +27,19 @@ from aleph.sdk.types import Account
 from aleph.sdk.chains.ethereum import get_fallback_account
 from aleph.sdk.conf import settings
 from aleph.sdk.vm.cache import VmCache
-from aleph_message.models import PostMessage
+from aleph_message.models import PostMessage, ItemHash, ChainRef
 
 from .utils import (
     subslices,
     async_iterator_to_list,
     IndexQuery,
     PageableResponse,
-    PageableRequest, EmptyAsyncIterator,
+    PageableRequest,
+    EmptyAsyncIterator,
 )
-from .exceptions import AlreadyForgottenError
+from .exceptions import AlreadyForgottenError, AlephPermissionError
 
+logger = logging.getLogger(__name__)
 R = TypeVar("R", bound="Record")
 
 
@@ -56,10 +59,12 @@ class Record(BaseModel, ABC):
     """
 
     forgotten: bool = False
-    id_hash: Optional[str] = None
+    id_hash: Optional[Union[ItemHash, str]] = None
     current_revision: Optional[int] = None
-    revision_hashes: List[str] = []
+    revision_hashes: List[ItemHash] = []
     timestamp: Optional[float] = None
+    signer: Optional[str] = None
+    __indexed_items: ClassVar[Set[str]] = set()
     __indices: ClassVar[Dict[str, "Index"]] = {}
 
     def __repr__(self):
@@ -68,13 +73,31 @@ class Record(BaseModel, ABC):
     def __str__(self):
         return f"{type(self).__name__} {self.__dict__}"
 
+    def __eq__(self, other):
+        return (
+            str(self.id_hash) == str(other.id_hash)
+            and self.current_revision == other.current_revision
+            and self.revision_hashes == other.revision_hashes
+            and self.forgotten == other.forgotten
+            and self.content == other.content
+            and self.signer == other.signer
+            # do not compare timestamps, they can deviate on Aleph between commitment and finalization
+        )
+
     @property
     def content(self) -> Dict[str, Any]:
         """
         :return: content dictionary of the object, as it is to be stored on Aleph.
         """
         return self.dict(
-            exclude={"id_hash", "current_revision", "revision_hashes", "forgotten", "timestamp"}
+            exclude={
+                "id_hash",
+                "current_revision",
+                "revision_hashes",
+                "forgotten",
+                "timestamp",
+                "signer",
+            }
         )
 
     async def update_revision_hashes(self: R):
@@ -112,10 +135,11 @@ class Record(BaseModel, ABC):
             else:
                 self.current_revision = rev_no
         elif rev_hash is not None:
+            rev_item_hash = ItemHash(rev_hash)
             if rev_hash == self.revision_hashes[self.current_revision]:
                 return self
             try:
-                self.current_revision = self.revision_hashes.index(rev_hash)
+                self.current_revision = self.revision_hashes.index(rev_item_hash)
             except ValueError:
                 raise IndexError(f"{rev_hash} is not a revision of {self}")
         else:
@@ -126,6 +150,7 @@ class Record(BaseModel, ABC):
         )
         self.__dict__.update(resp.content)
         self.timestamp = resp.timestamp
+        self.signer = resp.signer
 
         return self
 
@@ -142,6 +167,11 @@ class Record(BaseModel, ABC):
     def _index(self):
         for index in self.get_indices():
             index.add_record(self)
+        self.__indexed_items.add(self.id_hash)
+
+    @classmethod
+    def is_indexed(cls, id_hash: Union[ItemHash, str]):
+        return id_hash in cls.__indexed_items
 
     async def forget(self):
         """
@@ -149,6 +179,8 @@ class Record(BaseModel, ABC):
         The forgotten object should be deleted afterward, as it is useless now.
         """
         if not self.forgotten:
+            if self.signer != AARS.account.get_address():
+                raise AlephPermissionError(AARS.account.get_address(), self.id_hash, self.signer)
             await AARS.forget_objects([self])
             [index.remove_record(self) for index in self.get_indices()]
             self.forgotten = True
@@ -161,15 +193,21 @@ class Record(BaseModel, ABC):
         Initializes a record object from its PostMessage.
         :param post: the PostMessage to initialize from.
         """
-        obj = cls(**post.content.content)
+        obj: R = cls(**post.content.content)
         if post.content.ref is None:
             obj.id_hash = post.item_hash
         else:
-            obj.id_hash = post.content.ref
+            if isinstance(post.content.ref, str):
+                obj.id_hash = ItemHash(post.content.ref)
+            elif isinstance(post.content.ref, ChainRef):
+                obj.id_hash = post.content.ref.item_hash
+            else:
+                raise TypeError(f"Unknown type of ref: {type(post.content.ref)}")
         await obj.update_revision_hashes()
         assert obj.id_hash is not None
         obj.current_revision = obj.revision_hashes.index(obj.id_hash)
         obj.timestamp = post.time
+        obj.signer = post.sender
         return obj
 
     @classmethod
@@ -180,17 +218,20 @@ class Record(BaseModel, ABC):
         """
         obj = cls(**post["content"])
         if post.get("ref") is None:
-            obj.id_hash = post["item_hash"]
+            obj.id_hash = ItemHash(post["item_hash"])
         else:
-            obj.id_hash = post["ref"]
+            obj.id_hash = ItemHash(post["ref"])
         await obj.update_revision_hashes()
         assert obj.id_hash is not None
         obj.current_revision = obj.revision_hashes.index(obj.id_hash)
         obj.timestamp = post["time"]
+        obj.signer = post["sender"]
         return obj
 
     @classmethod
-    def fetch(cls: Type[R], hashes: Union[str, List[str]]) -> PageableResponse[R]:
+    def fetch(
+        cls: Type[R], hashes: Union[Union[str, ItemHash], List[Union[str, ItemHash]]]
+    ) -> PageableResponse[R]:
         """
         Fetches one or more objects of given type by its/their item_hash[es].
         """
@@ -248,6 +289,10 @@ class Record(BaseModel, ABC):
     def add_index(cls: Type[R], index: "Index") -> None:
         cls.__indices[repr(index)] = index
 
+    @classmethod
+    def remove_index(cls: Type[R], index: "Index") -> None:
+        del cls.__indices[repr(index)]
+
     @classmethod
     def get_index(cls: Type[R], index_name: str) -> "Index[R]":
         """
@@ -297,6 +342,35 @@ class Record(BaseModel, ABC):
             records.append(record)
         return records
 
+    @classmethod
+    async def forget_all(cls: Type[R]) -> List[ItemHash]:
+        """
+        #TODO: Check correctness
+        Forgets all Records of given type of the authorized user. If invoked on Record, will try to fetch all objects
+        of the current channel and forget them.
+
+        :return: The affected item_hashes
+        """
+        response = cls.fetch_objects()
+
+        item_hashes = []
+        record_batch = []
+        i = 0
+        async for record in response:
+            if record.signer != AARS.account.get_address():
+                continue
+            record_batch.append(record)
+            i += 1
+            if i % 50 == 0:
+                item_hashes.extend([record.id_hash for record in record_batch])
+                await AARS.forget_objects(record_batch)
+                record_batch = []
+        if record_batch:
+            item_hashes.extend([record.id_hash for record in record_batch])
+            await AARS.forget_objects(record_batch)
+
+        return item_hashes
+
 
 class Index(Record, Generic[R]):
     """
@@ -314,8 +388,8 @@ class Index(Record, Generic[R]):
         >>> Index(MyRecord, 'foo')
 
         This will create an index named 'MyRecord.foo', which is stored in the `MyRecord` class.
-        It is not recommended using the index directly, but rather through the `where_eq` method of the `Record` class like
-        so:
+        It is not recommended using the index directly, but rather through the `where_eq` method of the `Record` class
+        like so:
 
         >>> MyRecord.where_eq(foo='bar')
 
@@ -326,6 +400,10 @@ class Index(Record, Generic[R]):
         """
         if isinstance(on, str):
             on = [on]
+        # check if all properties exist
+        for prop in on:
+            if prop not in record_type.__fields__:
+                raise ValueError(f"Property {prop} does not exist on {record_type.__name__}")
         super(Index, self).__init__(record_type=record_type, index_on=sorted(on))
         record_type.add_index(self)
 
@@ -381,6 +459,8 @@ class Index(Record, Generic[R]):
         key = attrgetter(*self.index_on)(obj)
         if isinstance(key, str):
             key = (key,)
+        if isinstance(key, list):
+            key = tuple(key)
         if key not in self.hashmap:
             self.hashmap[key] = set()
         self.hashmap[key].add(obj.id_hash)
@@ -462,6 +542,10 @@ class AARS:
             channel = cls.channel
         assert isinstance(obj, Record)
         post_type = type(obj).__name__ if obj.id_hash is None else "amend"
+        if obj.id_hash is not None and obj.signer != cls.account.get_address():
+            raise AlephPermissionError(
+                obj.signer, obj.id_hash, cls.account.get_address()
+            )
         message, status = await cls.session.create_post(
             post_content=obj.content,
             post_type=post_type,
@@ -472,6 +556,8 @@ class AARS:
             obj.id_hash = message.item_hash
         obj.revision_hashes.append(message.item_hash)
         obj.current_revision = len(obj.revision_hashes) - 1
+        obj.timestamp = message.time
+        obj.signer = message.sender
         if cls.cache:
             await cls.cache.set(message.item_hash, obj.json())
         return obj
@@ -491,7 +577,12 @@ class AARS:
             channel = cls.channel
         hashes = []
         for obj in objs:
-            assert obj.id_hash is not None
+            if obj.id_hash is None:
+                raise ValueError("Cannot forget an object that has not been posted.")
+            if obj.signer != cls.account.get_address():
+                raise AlephPermissionError(
+                    obj.signer, obj.id_hash, cls.account.get_address()
+                )
             hashes += [obj.id_hash] + obj.revision_hashes
         forget_task = cls.session.forget(
             hashes=hashes,
@@ -507,7 +598,7 @@ class AARS:
     async def fetch_records(
         cls,
         record_type: Type[R],
-        item_hashes: Optional[List[str]] = None,
+        item_hashes: Optional[List[Union[str, ItemHash]]] = None,
         channel: Optional[str] = None,
         owner: Optional[str] = None,
         page_size: int = 50,
@@ -559,14 +650,14 @@ class AARS:
         return [
             record_type.parse_raw(r)
             for r in raw_records
-            if r is not None or not isinstance(r, BaseException)
+            if r is not None and not isinstance(r, BaseException)
         ]
 
     @classmethod
     async def _fetch_records_from_api(
         cls,
         record_type: Type[R],
-        item_hashes: Optional[List[str]] = None,
+        item_hashes: Optional[List[Union[str, ItemHash]]] = None,
         channels: Optional[List[str]] = None,
         owners: Optional[List[str]] = None,
         refs: Optional[List[str]] = None,
@@ -575,8 +666,11 @@ class AARS:
     ) -> AsyncIterator[R]:
         aleph_resp = None
         retries = cls.retry_count
+        if item_hashes is not None:
+            item_hashes = [str(h) for h in item_hashes]
         while aleph_resp is None:
             try:
+                actual_page = page if page != -1 else 1
                 aleph_resp = await cls.session.get_posts(
                     hashes=item_hashes,
                     channels=channels,
@@ -584,7 +678,7 @@ class AARS:
                     addresses=owners,
                     refs=refs,
                     pagination=page_size,
-                    page=page,
+                    page=actual_page,
                 )
             except ServerDisconnectedError:
                 retries -= 1
@@ -593,8 +687,8 @@ class AARS:
         for post in aleph_resp["posts"]:
             yield await record_type.from_dict(post)
 
-        if page == 1:
-            # If there are more pages, fetch them
+        if page == -1:
+            # Get all pages
             total_items = aleph_resp["pagination_total"]
             per_page = aleph_resp["pagination_per_page"]
             if total_items > per_page:
@@ -617,7 +711,7 @@ class AARS:
         channel: Optional[str] = None,
         owner: Optional[str] = None,
         page=1,
-    ) -> AsyncIterator[str]:
+    ) -> AsyncIterator[ItemHash]:
         """Retrieves posts of revisions of an object by its item_hash.
         :param record_type: The type of the objects to retrieve.
         :param ref: item_hash of the object, whose revisions to fetch.
@@ -638,6 +732,7 @@ class AARS:
                     addresses=owners,
                     refs=[ref],
                     pagination=50,
+                    page=page,
                 )
             except ServerDisconnectedError:
                 retries -= 1
@@ -659,7 +754,7 @@ class AARS:
                         owner=owner,
                         page=next_page,
                     ):
-                        yield item_hash
+                        yield ItemHash(item_hash)
 
     @classmethod
     async def fetch_exact(cls, record_type: Type[R], item_hash: str) -> R:
